--------------------------------------------------------------------------
-- Cpu_Times.Utilities                      Copyrigth ENS ULM 1998-1999 --
-- Pierre VINCENS                   V 01.000                 14/09/1998 --
--------------------------------------------------------------------------
-- VERSION Unix pour Linux
--------------------------------------------------------------------------
with System;
with Interfaces.C; use type Interfaces.C.Long;

with Ada.Text_Io; use Ada.Text_Io;
with Ada.Integer_Text_Io; use Ada.Integer_Text_Io;
with Ada.Calendar; use Ada.Calendar;

with Strlib; use Strlib;

with System_Constants; use System_Constants;

with Cpu_Times.Basic; 

package body Cpu_Times.Utilities is


  package Duration_IO is new Ada.Text_Io.Fixed_Io(Duration);
  use Duration_IO;

----------------------------------------------------------------------------------
-- Fonctions de services pour determiner les temps consommes pour un process
-- Cette partie est specifique linux
----------------------------------------------------------------------------------

  No_Pid : exception;
  -- Lev si un numro de PID demand n'existe plus

  type Cpu_Times_Data is record
    U_Time  : Integer;
    S_Time  : Integer;
    Cu_Time : Integer;
    Cs_Time : Integer;
  end record;

  function Get_Times_From_Stat(Pid : in Process_Id) return Cpu_Times_Data is
    -- Recuperation de l'information  partir du fichier /proc/<pid>/stat
    Spid  : constant String := Process_Id'Image(Pid); 
    Fname : constant String := "/proc/" & Spid(Spid'First+1..Spid'Last) & "/stat";
    Fic   : Ada.Text_Io.File_Type;
    Line  : String(1..2024);
    Last  : Natural;
    Key   : Varstring;
    Position : Integer := Line'First;
    Elapsed : Cpu_Times_Data;
  begin
    -- Lecture des infos dans le fichier
    Open(Fic,In_File,Fname);
    Get_Line(Fic,Line,Last);
    Close(Fic);

    -- Dcomposition de la ligne
    -- Lecture du numero du process
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture du nom du process
    Next_Word(Line(Position..Last),Key,Position,")");
    -- On saute l'etat du process (1 caractere)
    Position := Position + 3;
    -- Lecture du ppid
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture du pgrp
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture de session
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture du tty
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture du tpgid
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture des flags
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture des flags minflt
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture des flags cminflt
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture des flags majflt
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture des flags cmajflt
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture des flags utime
    Next_Word(Line(Position..Last),Key,Position," ");
    Elapsed.U_Time := Integer'Value(Key.Value.all);
    -- Lecture des flags stime
    Next_Word(Line(Position..Last),Key,Position," ");
    Elapsed.S_Time := Integer'Value(Key.Value.all);
    -- Lecture des flags cutime
    Next_Word(Line(Position..Last),Key,Position," ");
    Elapsed.Cu_Time := Integer'Value(Key.Value.all);
    -- Lecture des flags cstime
    Next_Word(Line(Position..Last),Key,Position," ");
    Elapsed.Cs_Time := Integer'Value(Key.Value.all);
    -- On s'arrete l
    return Elapsed;
  exception
    when Name_Error | Device_Error => raise No_Pid;
  end Get_Times_From_Stat;

  function Get_Times(Pid : in Process_Id) return Cpu_Times_Data is
    -- Recuperation de l'information  partir du fichier /proc/<pid>/cpu
    Spid  : constant String := Process_Id'Image(Pid); 
    Fname : constant String := "/proc/" & Spid(Spid'First+1..Spid'Last) & "/cpu";
    Fic   : Ada.Text_Io.File_Type;
    Line  : String(1..2024);
    Last  : Natural;
    Key   : Varstring;
    Position : Integer := Line'First;
    Elapsed : Cpu_Times_Data;
  begin
    -- Lecture des infos dans le fichier
    Open(Fic,In_File,Fname);
    Get_Line(Fic,Line,Last);
    Close(Fic);

    -- Dcomposition de la ligne
    -- Lecture du mot cl cpu
    Next_Word(Line(Position..Last),Key,Position," ");
    -- Lecture des flags utime
    Next_Word(Line(Position..Last),Key,Position," ");
    Elapsed.U_Time := Integer'Value(Key.Value.all);
    -- Lecture des flags stime
    Next_Word(Line(Position..Last),Key,Position," ");
    Elapsed.S_Time := Integer'Value(Key.Value.all);
    return Elapsed;
  exception
    when Name_Error => 
      -- Pas de fichier /proc/pid/cpu. On essaie /proc/pid/stat.
      return Get_Times_From_Stat(Pid);
  end Get_Times;

----------------------------------------------------------------------------------
-- Gestion des listes de process
----------------------------------------------------------------------------------
  use Process_Unit_List;

  function Equal_Pid(Pu  : in Process_Unit;
		     Pid : in Process_Id) return Boolean;
  pragma Inline(Equal_Pid);
  function Equal_Pid(Pu  : in Process_Unit;
		     Pid : in Process_Id) return Boolean is
  begin
    return Pid = Pu.Pid;
  end Equal_Pid;
  function Find_Pid_In_List is new 
    Process_Unit_List.Find_Element(Process_Id,Equal_Pid);

  procedure Min_Max(List : in Process_List; Min, Max : in out Process_Id) is
    procedure Check(Pu  : in out Process_Unit) is
    begin
      Min := Process_Id'Min(Min,Pu.Pid);
      Max := Process_Id'Max(Max,Pu.Pid);
    end Check;
    procedure Check_All is new Process_Unit_List.Apply(Check);
  begin
    Min := Process_Id'Last;
    Max := Process_Id'First;
    Check_All(List);
  end Min_Max;

  procedure Add_Pid_In_List(List : in out Process_List;
			    Key  : in String) is
    Pid : constant Process_Id := Getpid;
    Pu  : Process_Unit;
    Adr : Process_Unit_List.List_Ptr := Find_Pid_In_List(List,Pid);
  begin
    if Adr = Nil then
      Pu.Pid := Pid;
      Pu.Key := To_Varstring(Key);
      Append(List,Pu);
    end if;
  end Add_Pid_In_List;

  procedure Remove_Pid_From_List(List : in out Process_List) is
    Pid : constant Process_Id := Getpid;
    Adr : Process_Unit_List.List_Ptr := Find_Pid_In_List(List,Pid);
    Pu  : Process_Unit; 
    T_Elapsed : Cpu_Times_Data;
  begin
    if Adr /= Nil then
      Pu := Consult(Adr);
      -- Annulation de l'adresse
      Pu.Up := False;
      -- Enregistrement du dernier temps comptabilise
      begin
	T_Elapsed := Get_Times(Pu.Pid);
	Pu.User_Time := Duration(T_Elapsed.U_Time);
	Pu.System_Time := Duration(T_Elapsed.S_Time);
      exception
	when No_Pid =>
	  null;
      end;
      Replace(Adr,Pu);
    end if;
  end Remove_Pid_From_List;

  procedure Add_Pid_In_List(List : in out Process_List;
			    Pid  : in Process_Id;
			    Key  : in String) is
    Adr : Process_Unit_List.List_Ptr := Find_Pid_In_List(List,Pid);
    Pu  : Process_Unit;
  begin 
    if Adr = Nil then 
      Pu.Pid := Pid;
      Pu.Key := To_Varstring(Key);
      Append(List,Pu);
    end if;
  end Add_Pid_In_List;

  procedure Remove_Pid_From_List(List : in out Process_List;
				 Pid  : in Process_Id) is
    Adr : Process_Unit_List.List_Ptr := Find_Pid_In_List(List,Pid);
    Pu  : Process_Unit; 
    T_Elapsed : Cpu_Times_Data;
  begin
    if Adr /= Nil then
      Pu := Consult(Adr);
      -- Annulation de l'adresse
      Pu.Up := False;
      -- Enregistrement du dernier temps comptabilise
      begin
	T_Elapsed := Get_Times(Pu.Pid);
	Pu.User_Time := Duration(T_Elapsed.U_Time);
	Pu.System_Time := Duration(T_Elapsed.S_Time);
      exception
	when No_Pid =>
	  null;
      end;
      Replace(Adr,Pu);
    end if;
  end Remove_Pid_From_List;

  procedure Add_Childs(List : in out Process_List) is
    -- Recuperation de l'information  partir du rpertoire /proc
    function Proc_Name(Pid : in Process_Id) return String;
    pragma Inline(Proc_Name);
    function Proc_Name(Pid : in Process_Id) return String is
      Spid  : constant String := Process_Id'Image(Pid); 
      Fname : constant String := 
	"/proc/" & Spid(Spid'First+1..Spid'Last) & "/stat";
    begin
      return Fname;
    end Proc_Name;
    Fic   : Ada.Text_Io.File_Type;
    Pid_Min,Pid_Max : Process_Id;
  begin
    Min_Max(List,Pid_Min,Pid_Max);
    Pid_Max := Pid_Max + 30; -- Pour se donner de la marge
    -- On teste l'existence des process 
    for I in Pid_Min..Pid_Max loop
      declare
	Pname : constant String := Proc_Name(I);
	Line  : String(1..2024);
	Last  : Natural;
	Key   : Varstring;
	Position : Integer := Line'First;
	Ppid  : Process_Id;
	Adr : Process_Unit_List.List_Ptr := Find_Pid_In_List(List,I);
      begin
	-- Le process a teste est il deja dans la liste
	if Adr = Nil then
	  -- Non, il faut donc le tester
	  -- Lecture des infos dans le fichier
	  Open(Fic,In_File,Pname);
	  Get_Line(Fic,Line,Last);
	  Close(Fic);
	  -- Dcomposition de la ligne
	  -- Lecture du numero du process
	  Next_Word(Line(Position..Last),Key,Position," ");
	  -- Lecture du nom du process
	  Next_Word(Line(Position..Last),Key,Position,")");
	  -- On saute l'etat du process (1 caractere)
	  Position := Position + 3;
	  -- Lecture du ppid
	  Next_Word(Line(Position..Last),Key,Position," ");
	  -- Transformation de la chaine
	  Ppid := Process_Id'Value(To_String(Key));
	  -- Le process pre est-il dans la liste?
	  Adr := Find_Pid_In_List(List,Ppid);
	  if Adr /= Nil then
	    -- Oui, donc il faut ajouter le fils
	    Add_Pid_In_List(List,I,"System");
	  end if;
	end if;
      exception
	when Name_Error | Device_Error => null;
      end;
    end loop;
  end Add_Childs; 

  procedure Print_Process_List(List : in Process_List) is
    procedure Print_One_Process(Pu : in out Process_Unit) is
    begin
      Put("   Pid:"); Put(Integer(Pu.Pid),8);
      if Pu.Up then Put(" Up  "); else Put(" Down"); end if;
      Put(" "); 
      Put(Pu.User_Time/Clk_Tck,8,2); 
      Put("u ");
      Put(Pu.System_Time/Clk_Tck,8,2); 
      Put("s ");
      Put_line("Key: " & To_String(Pu.Key));
    end Print_One_Process;
    procedure Print_Processes is new Apply(Print_One_Process);
  begin
    Put_Line("List of Processes:");
    Print_Processes(List);
  end Print_Process_List;

  procedure Print_Elapsed_Times(List : in Process_List) is
    U_Time,S_Time : Duration := 0.0;
    Elapsed : constant Duration := Cpu_Times.Basic.Elapsed_Real_Time;
    Hours   : constant Integer  := Integer(Elapsed)/3600;
    Minutes : constant Integer  := (Integer(Elapsed)-Hours*3600)/60;
    Seconds : constant Duration := 
      abs(Elapsed - Duration(Hours*3600+Minutes*60));
    procedure Add_One_Process(Pu : in out Process_Unit) is
    begin
      U_Time := U_Time + Pu.User_Time;
      S_Time := S_Time + Pu.System_Time;
    end Add_One_Process;
    procedure Sum_Of_Processes is new Apply(Add_One_Process);
  begin
    Sum_Of_Processes(List);
    Put("Elapsed Times: ");
    Put(U_Time/Clk_Tck,0,2); Put("u ");
    Put(S_Time/Clk_Tck,0,2); Put("s ");
    Put((U_Time+S_Time)/Clk_Tck,0,2); Put("t ");
    Put(Elapsed,0,2); Put("r ");
    Put(Hours,0);
    Put(':');
    Put(Minutes,0);
    Put(':');
    Put(Seconds,0,2);
    New_Line;
  end Print_Elapsed_Times;

----------------------------------------------------------------------------------
-- Gestion des temps
----------------------------------------------------------------------------------

  function Elapsed_Times(Pid : in Process_Id) return Cpu_Duration is
    T_Elapsed : constant Cpu_Times_Data := Get_Times(Pid);
    Elapsed : Cpu_Duration;
  begin
    Elapsed.Real_Time := Cpu_Times.Basic.Elapsed_Real_Time;
    Elapsed.User_Time := Duration(T_Elapsed.U_Time)/Clk_Tck;
    Elapsed.System_Time := Duration(T_Elapsed.S_Time)/Clk_Tck;
    return Elapsed;
  end Elapsed_Times;
 
  procedure Elapsed_Times(List : in Process_List; 
			  Elapsed : in out Cpu_Duration) is
    -- A chaque appel, on cumule les temps afin de tenir compte
    -- des process termines
    procedure Sum_Times_For_One_Process(Pu : in out Process_Unit) is
      T_Elapsed : Cpu_Times_Data;
    begin
      if Pu.Up then
	begin
	  -- Enregistrement du dernier temps comptabilise
	  T_Elapsed := Get_Times(Pu.Pid);
	  Pu.User_Time := Duration(T_Elapsed.U_Time);
	  Pu.System_Time := Duration(T_Elapsed.S_Time);
	exception
	  when No_Pid => -- Le process n'existe plus
	    Pu.Up := False;
	end;
      end if;
      -- Comptabilisation du temps
      Elapsed.User_Time := Elapsed.User_Time + Pu.User_Time;
      Elapsed.System_Time := Elapsed.System_Time + Pu.System_Time;    
    end Sum_Times_For_One_Process;
    procedure Sum_Times is new Apply(Sum_Times_For_One_Process);
  begin
    Elapsed := Zero;
    Elapsed.Real_Time := Cpu_Times.Basic.Elapsed_Real_Time;
    Sum_Times(List);
    Elapsed.User_Time := Elapsed.User_Time/Clk_Tck;
    Elapsed.System_Time := Elapsed.System_Time/Clk_Tck;
  end Elapsed_Times;

end Cpu_Times.Utilities;
