--------------------------------------------------------------------------
-- Dassirc_Builder                          Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 01.000                 13/03/1998 --
--------------------------------------------------------------------------
-- Cette procedure cliente a la charge de rechercher les squences 
-- similaires pour les valeurs de masques de graines qui lui sont
-- affecte par le dispatcher
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Calendar;
with Ada.Exceptions; use Ada.Exceptions;

with System_Constants; use System_Constants;
with System_Functions; use System_Functions;

with Type_Specifications; use Type_Specifications;
with Cpu_Times; use Cpu_Times;
with Cpu_Times.Basic; use Cpu_Times.Basic;
with Cpu_Times.Counters; use Cpu_Times.Counters;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;

with Sequence_Alignment; use Sequence_Alignment;

with Assirc_Specifications; use Assirc_Specifications;
with Assirc_Specifications.Basic; use Assirc_Specifications.Basic;
with Assirc_Specifications.Database; use Assirc_Specifications.Database;
with Assirc_Seeds; use Assirc_Seeds;
with Assirc_Seeds.Database;  -- Visibilit dans les taches correspondantes
with Assirc_Seeds.D_Eval_M1; --             ''
with Assirc_Seeds.D_Eval_M2; --             ''
with Assirc_Seeds.D_Eval_M3; --             ''
with Assirc_Seeds.D_Eval_M4; --             ''
with Assirc_Statistics;      --             ''

with Dassirc_Common; use Dassirc_Common;
with Dassirc_Alignment; use Dassirc_Alignment;
with Dassirc_Builder_Options; use Dassirc_Builder_Options;
with Dassirc_Database;   -- Pas de visibilit explicite
with Dassirc_Dispatcher; --            ''
with Dassirc_Messages; use Dassirc_Messages;
with Dassirc_Utils; use Dassirc_Utils;
with Dassirc_Semaphore; use Dassirc_Semaphore;
with Dassirc_Statistics;

procedure Dassirc_Builder is
  Client        : View_Id := Unassigned_Id;
  Max_Sequences : constant Accession_Type := 2;
  Sequences     : Assirc_Sequence_Array_Access;
  Parameters    : Searching_Parameters;
  Options       : Builder_Options;
  
  Client_Aborted : exception;

--------------------------------------------------------------------------
-- Un type protg pour comptabiliser le nombre de taches actives       --
--------------------------------------------------------------------------
  type Builder_Status is array(Action_Type) of Status_Type;
  protected Work is
    procedure Started(Action : in Action_Type);
    -- Pour indiquer qu'une tache vient de s'activer (etat waiting)
    procedure Run(Action : in Action_Type); 
    -- Pour indiquer qu'une tache a t ractive (waiting -> running)
    procedure Wait(Action : in Action_Type);
    -- Pour indiquer qu'une tache est en attente (running -> waiting)
    procedure Terminated(Action : in Action_Type);
    -- Pour indiquer qu'une tache vient de se terminer (doit etre dans
    -- l'etat waiting) (waiting -> terminated);
    function Get_Status return Status_Type;
    -- Pour obtenir le status de la tache
    entry End_For_Searching;
    -- Est bloquant tant que le nombre de taches actives pour la recherche
    -- des paires hypothtiques n'est pas nul;
    entry Nothing;
    -- Est bloquant tant que le nombre de taches actives n'est pas nul;
    function Non_Active_Task return Boolean;
    -- Indique s'il n'y a plus de tache active;
  private
    Count : Natural := 0;
    Count_For_Searching  : Natural := 0;
    Count_Expand_Waiting : Natural := 0; 
    Count_Expand_Running : Natural := 0; 
    Updated : Builder_Status := ((others => Unknown));  -- Etat local
  end Work;

  protected body Work is
    procedure Started(Action : in Action_Type) is
    begin
      Count := Count + 1;
      case Action is
	when Region_Computing =>
	  Count_For_Searching := Count_For_Searching + 1;
	  Updated(Action) := Waiting;
	when Seed_Computing =>
	  Count_For_Searching := Count_For_Searching + 1;
	  Updated(Action) := Waiting;
	when Expand_Computing =>
	  Count_For_Searching := Count_For_Searching + 1;
	  if Count_Expand_Running = 0 and Count_Expand_Waiting = 0 then
	    Updated(Action) := Waiting;
	  end if;
	  Count_Expand_Waiting := Count_Expand_Waiting + 1;
	when Alignment_Computing =>
	  Updated(Action) := Waiting;
      end case;
    end Started;
    procedure Run(Action : in Action_Type) is
    begin
      if Action = Expand_Computing then
        if Updated(Action) = Waiting and Count_Expand_Running = 0 then
          Updated(Action) := Running;
        end if;
        Count_Expand_Waiting := Count_Expand_Waiting - 1; 
        Count_Expand_Running := Count_Expand_Running + 1; 
      else
        Updated(Action) := Running;
      end if;
    end Run;
    procedure Wait(Action : in Action_Type) is
    begin
      if Action = Expand_Computing then
        if Count_Expand_Running = 1 then
          Updated(Action) := Waiting;
        end if;
        Count_Expand_Waiting := Count_Expand_Waiting + 1; 
        Count_Expand_Running := Count_Expand_Running - 1; 
      else
        Updated(Action) := Waiting;
      end if;
    end Wait;
    procedure Terminated(Action : in Action_Type) is
    begin
      Count := Count - 1;
      case Action is
	when Region_Computing =>
	  Count_For_Searching := Count_For_Searching - 1;
	  Updated(Action) := Terminated;
	when Seed_Computing =>
	  Count_For_Searching := Count_For_Searching - 1;
	  Updated(Action) := Terminated;
	when Expand_Computing =>
	  Count_For_Searching := Count_For_Searching - 1;
	  if Count_Expand_Waiting = 1 then
	    Updated(Action) := Terminated;
	  end if;
	  Count_Expand_Waiting := Count_Expand_Waiting - 1; 
	when Alignment_Computing =>
	  Updated(Action) := Terminated;
      end case;
    end Terminated; 
    function Get_Status return Status_Type is
      Status : Status_Type := Unknown;
    begin
      for I in Updated'Range loop
        if Updated(I) = Running  then
          return Running;
        elsif Updated(I) = Waiting then
          Status := Waiting;
        elsif Updated(I) = Terminated and Status = Unknown then
          Status := Terminated;
        end if;
      end loop;
      return Status;
    end Get_Status;
    entry End_For_Searching when Count_For_Searching = 0 is
    begin
      null;
    end End_For_Searching;
    entry Nothing when Count = 0 is
    begin
      null;
    end Nothing;
    function Non_Active_Task return Boolean is
    begin
      return Count = 0;
    end Non_Active_Task;   
  end Work;

--------------------------------------------------------------------------
-- La tache de gnration des graines                                   --
--------------------------------------------------------------------------
  task Seed_Computing_Task is
    -- Cette tache prend en charge la gnration des graines pour la
    -- recherche. Elle n'effectue pas le calcul de l'expansion
    entry Start(Method : in Method_Of_Seed_Computing);    -- Activer la tache
    entry Shutdown; -- Desactiver cette tache 
  end Seed_Computing_Task;

  task body Seed_Computing_Task is
    use Assirc_Seeds.D_Eval_M2;
    use Assirc_Seeds.D_Eval_M3;
    Mask    : Integer;
    Seeds   : Seed_Segments;
    Computing_Method : Method_Of_Seed_Computing;
    NO_TREATMENT : exception;
  begin
    Cpu_Counters.Add("Seed_Computing_Task");
    Put_Line("Initialisation de la tache: Seed_Computing_Task");       
    Work.Started(Seed_Computing);
    select
      accept Start(Method : in Method_Of_Seed_Computing) do
        Computing_Method := Method;
      end Start;
    or
      accept Shutdown;
      raise NO_TREATMENT;
    or
      terminate;
    end select;
    Put_Line("Activation de la tache: Seed_Computing_Task");
    case Computing_Method is
      when Simple =>
	-- Si la partition ne fait que de la production de graines, il faut
	-- initialiser les couples de brins a analyser.
	Assirc_Seeds.D_Eval_M3.Set_Pairs_To_Check(Sequences);
	-- et arreter la tache rcuprant les mises a jour de la base
	Assirc_Seeds.D_Eval_M4.Shutdown_Local_Database;
      when Combined =>
	-- Arreter la tache rcuprant les mises a jour de la base
	Assirc_Seeds.D_Eval_M4.Shutdown_Local_Database;
      when Local_Combined =>
	-- Activer la tache recuprant les mise  jour de la base
	Put_Line("Seed_Computing_Task: initialisation de la base locale");
	Assirc_Seeds.D_Eval_M4.Start_Local_Database(Client);
	Put_Line("Seed_Computing_Task: base locale initialisee");
    end case;       
    Put_Line("Initialisation de l'espace memoire ncessaire aux graines");
    Init_Seeds(Parameters.Seed_Size-Parameters.Mask_Size,Seeds);
    Put_Line("Espace memoire initialis");
    loop 
      -- on sort de cette boucle et donc de la tache sur l'exception 
      -- NO_MASK_AVAILABLE
      Put_Line("Demande du prochain masque");
      Dispatcher_Access.Lock;
      Mask := Dassirc_Dispatcher.Get_Seed_Mask(Client);
      Dispatcher_Access.Unlock;
      if Mask < 0 then
	-- a cause du bug linux
	raise NO_MASK_AVAILABLE;
      end if;
      Put_Line("Traite le masque " & Integer'Image(Mask));
      Work.Run(Seed_Computing);
      Put_Line("Run Seed Computing");
      for I in Sequences'Range loop
        Put_Line("Gnration des graines pour le masque " & 
                 Integer'Image(Mask) & " sur la squence " & 
                 Accession_Type'Image(I));
        if Sequences(I).Strand = Watson or Sequences(I).Strand = Both then
          Get_Seeds_By_Mask(Sequences(I),Watson,
                            Parameters.Seed_Size,Parameters.Mask_Size,Mask,
                            Seeds);
        end if;
        if Sequences(I).Strand = Crick or Sequences(I).Strand = Both then
          Get_Seeds_By_Mask(Sequences(I),Crick,
                            Parameters.Seed_Size,Parameters.Mask_Size,Mask,
                            Seeds);
        end if;
        Put_Line("Graines gnres pour le masque " & Integer'Image(Mask) &
                 " sur la squence " & Accession_Type'Image(I));
      end loop;
      Put_Line("Evaluation des graines pour le masque " & 
               Integer'Image(Mask) &
	       " (Method: " & 
	       Method_Of_Seed_Computing'Image(Computing_Method) & ")");
      case Computing_Method is
        when Simple =>
          Assirc_Seeds.D_Eval_M3.Product_Seeds
	    (Seeds,
	     Parameters.Common_Max_Dist,
	     Parameters.Xonly);
        when Combined =>
          Assirc_Seeds.D_Eval_M2.Eval_Seeds
	    (Sequences,Seeds,
	     Parameters.Recorded_Length,Parameters.Distance,
	     Parameters.Tolerance,Parameters.Common_Max_Dist,
	     Parameters.Skip,Parameters.Xonly);
        when Local_Combined =>
	  Put_Line("Evaluation des graines: modele 4");
          Assirc_Seeds.D_Eval_M4.Eval_Seeds
	    (Sequences,Seeds,
	     Parameters.Recorded_Length,Parameters.Distance,
	     Parameters.Tolerance,Parameters.Common_Max_Dist,
	     Parameters.Skip,Parameters.Xonly);
      end case;
      Put_Line("Graines values pour le masque " & 
                       Integer'Image(Mask));
      Reset_Seeds(Seeds);
      Put_Line("Mmoire rinitialise."); 
      Dispatcher_Access.Lock;
      Dassirc_Dispatcher.Mask_Completed(Client,Mask);
      Dispatcher_Access.Unlock;
      Put_Line("Masque complt."); 
      Work.Wait(Seed_Computing);
    end loop;
    if Computing_Method = Local_Combined then
      -- En mode local_combined, on informe la base remote que c'est fini
      Assirc_Seeds.D_Eval_M4.Stop_Local_Database;
    end if;
    -- On dtruit la base de donnes locale
    Assirc_Seeds.Database.Close_Local_Database; 
    Cpu_Counters.Remove;
    Put_Line("Terminaison Anormale de la tache: Seed_Computing_Task");
  exception
    when NO_MASK_AVAILABLE =>
      Put_Line("Fin de traitement des masques");
      Free_Seeds(Seeds);
      case Computing_Method is
        when Simple   => 
	  Assirc_Seeds.D_Eval_M3.Reset_Pairs_To_Check; 
	when Combined => 
	  Assirc_Seeds.D_Eval_M2.Reset_Pairs_To_Check;
      	when Local_Combined => 
	  -- En mode local_combined, on informe la base remote que c'est fini
	  Assirc_Seeds.D_Eval_M4.Stop_Local_Database;
	  Assirc_Seeds.D_Eval_M2.Reset_Pairs_To_Check;
      end case;
      Assirc_Seeds.Database.Close_Local_Database;
      Work.Terminated(Seed_Computing);
      Cpu_Counters.Remove;
      Put_Line("Terminaison de la tache: Seed_Computing_Task");       
    when NO_TREATMENT =>
      Put_Line("Dsactivation de la tache de construction des graines");
      -- Arreter la tache rcuprant les mises a jour de la base
      Assirc_Seeds.D_Eval_M4.Shutdown_Local_Database;
      -- Informer que cette tache est termine
      Work.Terminated(Seed_Computing);
      Cpu_Counters.Remove;
      Put_Line("Terminaison de la tache: Seed_Computing_Task");       
    when The_Error : others =>
      Put_Line("Exception raised in Seed_Computing_Task");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      Work.Wait(Seed_Computing);
      Work.Terminated(Seed_Computing);
      Cpu_Counters.Remove;
  end Seed_Computing_Task;
  
--------------------------------------------------------------------------
-- La tache d'expansion des graines                                     --
--------------------------------------------------------------------------
  task type Expand_Computing_Task is
    entry Start;
  end Expand_Computing_Task;
  type Expand_Computing_Array is array(Integer range <>) of Expand_Computing_Task;
  type Expand_Computing_Access is access Expand_Computing_Array;

  task body Expand_Computing_Task is
    use Assirc_Seeds.D_Eval_M3;
  begin
    Cpu_Counters.Add("Expand_Computing_Task");
    Put_Line("Initialisation de la tache: Expand_Computing_Task");       
    Work.Started(Expand_Computing);
    -- Le rendez-vous start ne sert qu'a garantir l'initialisation
    -- avant que le programme principal ne se mette en attente
    -- de la fin du travail de recherche.
    accept Start; 
    Put_Line("Activation de la tache: Expand_Computing_Task");       
    Work.Run(Expand_Computing);
    Expand_Seeds(Sequences,
                 Parameters.Recorded_Length,Parameters.Distance,
                 Parameters.Tolerance,Parameters.Common_Max_Dist,
                 Parameters.Skip);
    Work.Wait(Expand_Computing);
    Work.Terminated(Expand_Computing);
    Cpu_Counters.Remove;
    Put_Line("Terminaison de la tache: Expand_Computing_Task");     
  exception
    when The_Error : others =>
      Put_Line("Exception raised in Expand_Computing_Task");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      Work.Wait(Expand_Computing);
      Work.Terminated(Expand_Computing);
      Cpu_Counters.Remove;
  end Expand_Computing_Task;

  Run_Of_Expand_Computing : Expand_Computing_Access;

--------------------------------------------------------------------------
-- La tache de gnration des graines sur une zone limite              --
--------------------------------------------------------------------------
-- Cette tache prend en compte le calcul complet des graines et de leur
-- extension, mais pour une rgion limite.
-- Une base de donnes interne est gre.
-- Les rsultats sont transmis  la base commune  la fin du calcul
--------------------------------------------------------------------------
  task Region_Computing_Task is
    -- Cette tache prend en charge la gnration des graines pour la
    -- recherche. Elle n'effectue pas le calcul de l'expansion
    entry Start;    -- Activer la tache
    entry Shutdown; -- Desactiver cette tache 
  end Region_Computing_Task;

  task body Region_Computing_Task is
    use Assirc_Seeds.D_Eval_M1;
    Seeds    : Seed_Segments;
    NO_TREATMENT : exception;
    Z1_Xonly,Z2_Xonly : Boolean;
    Zone1,Zone2 : Segment_Type;
    
    procedure Compute_Seeds(Id     : in Accession_Type;
                            Strand : in Strand_Type) is
    begin
      Put_Line("Gnration des graines sur la squence " & 
                Accession_Type'Image(Id));
      if Strand = Watson or Strand = Both then
        Get_Seeds(Sequences(Id),Watson,Parameters.Seed_Size,Seeds);
      end if;
      if Strand = Crick or Strand = Both then
        Get_Seeds(Sequences(Id),Crick,Parameters.Seed_Size,Seeds);
      end if;
      Put_Line("Graines gnres sur la squence " & 
               Accession_Type'Image(Id));
    end Compute_Seeds;
    
    procedure Compute_Seeds_By_Mask(Id     : in Accession_Type;
                                    Strand : in Strand_Type;
				    Mask   : in Integer) is
    begin
      Put_Line("Gnration des graines pour le masque " & 
               Integer'Image(Mask) & " sur la squence " & 
               Accession_Type'Image(Id));
      if Strand = Watson or Strand = Both then
        Get_Seeds_By_Mask(Sequences(Id),Watson,
                          Parameters.Seed_Size,Parameters.Mask_Size,Mask,
                          Seeds);
      end if;
      if Strand = Crick or Strand = Both then
        Get_Seeds_By_Mask(Sequences(Id),Crick,
                          Parameters.Seed_Size,Parameters.Mask_Size,Mask,
                          Seeds);
      end if;
      Put_Line("Graines gnres pour le masque " & Integer'Image(Mask) &
               " sur la squence " & Accession_Type'Image(Id));
    end Compute_Seeds_By_Mask;
    
  begin  
    Cpu_Counters.Add("Region_Computing_Task");
    Put_Line("Activation de la tache: Region_Computing_Task");   
    Work.Started(Region_Computing);
    select
      accept Start;
    or
      accept Shutdown;
      raise NO_TREATMENT;
    or
      terminate;
    end select;
    Put_Line("Calcul par rgion");
    Put_Line("Initialisation de la base de donnes locale");
    Assirc_Seeds.Database.Set_Parameters(Parameters.Excluded_Marge,
					 Parameters.Intersection_Marge);
    Put_Line("Parameters.Excluded_Marge: " & 
	     Integer'Image(Parameters.Excluded_Marge));
    Put_Line("Parameters.Intersection_Marge: " & 
	     Integer'Image(Parameters.Intersection_Marge));
    Put_Line("Initialisation de l'espace memoire ncessaire aux graines");
    Init_Seeds(Parameters.Seed_Size-Parameters.Mask_Size,Seeds);
    Put_Line("Espace memoire initialis");
    Put_Line("Parameters.Seed_Size: " & Integer'Image(Parameters.Seed_Size));
    Put_Line("Parameters.Mask_Size: " & Integer'Image(Parameters.Mask_Size));
    loop 
      -- On sort de la boucle sur l'exception NO_REGION_AVAILABLE du 
      -- dispatcher. Le calcul se fait sur une zone limite de la premire 
      -- squence.
      -- La demander au dispatcher
      Put_Line("Demande de la prochaine rgion");
      Dispatcher_Access.Lock;
      Dassirc_Dispatcher.Get_Region(Client,
				    Zone1,Zone2,
				    Z1_Xonly,Z2_Xonly);
      Dispatcher_Access.Unlock;
      if Zone1.Start = Integer'First then
	-- Ce test permet de recuperer l'exception qui devrait etre transmise
	-- par dassirc_manager. Il semblerait qu'il y ait un bug de ce cote
	-- l d'o ce patch.
	raise NO_REGION_AVAILABLE;
      end if;
      -- Definir sur quel brin doit se faire la recherche
      Zone1.Strand := Sequences(Zone1.Id).Strand;
      if Zone2.Id >= Sequences'First and Zone2.Id <= Sequences'Last then
        Zone2.Strand := Sequences(Zone2.Id).Strand;
        --if Zone1.Strand = Both and Zone2.Strand = Both then
	  -- Dans ce cas, compte-tenu de la symtrie, il suffit de considrer
	  -- un seul brin pour la premire squence
          --Zone1.Strand := Watson;
        --end if;
      end if;
      Work.Run(Region_Computing);
      Put_Line("Traite la sequence" & Accession_Type'Image(Zone1.Id) &
               " de " & Integer'Image(Zone1.Start) & 
	       "  " & Integer'Image(Zone1.Stop));
      if Zone2.Id >= Sequences'First and Zone2.Id <= Sequences'Last then
        Put_Line("avec la sequence" & Accession_Type'Image(Zone2.Id) &
		 " de " & Integer'Image(Zone2.Start) & 
		 "  " & Integer'Image(Zone2.Stop));
      end if;
      Put_Line("avec Z1_Xonly = " & Boolean'Image(Z1_Xonly) &
	       " , Z2_Xonly = " & Boolean'Image(Z2_Xonly) &
	       " et Parameters.Xonly = " & Boolean'Image(Parameters.Xonly));
      if Parameters.Mask_Size = 0 then
        -- Il n'y a pas de masque
	-- On positionne les bornes pour la premiere sequence
	Sequences(Zone1.Id).Start := Zone1.Start;
	Sequences(Zone1.Id).Stop  := Zone1.Stop;
        Put_Line("Calcul des graines");
	Compute_Seeds(Zone1.Id,Zone1.Strand);
        if Zone2.Id in Sequences'Range then
          -- On positionne les bornes pour la seconde squence
	  -- cela se fait ici, car id2 peut etre egal a id1
	  Sequences(Zone2.Id).Start := Zone2.Start;
	  Sequences(Zone2.Id).Stop  := Zone2.Stop;
	  Compute_Seeds(Zone2.Id,Zone2.Strand);
        end if;
	-- Eval_Seeds n'utilise pas les positions sequences(id).start
	-- ou stop. Si cela tait modifie, il faut revoir cette partie
	-- toujours a cause du fait que Id1 peut etre egal a Id2
        Put_Line("Evaluation des graines");
        Eval_Seeds(Sequences,Seeds,
                   Parameters.Recorded_Length,Parameters.Distance,
                   Parameters.Tolerance,Parameters.Common_Max_Dist,
                   Parameters.Skip,Parameters.Xonly,
                   Zone1,Zone2,Z1_Xonly,Z2_Xonly);
        Put_Line("Graines values");
        Reset_Seeds(Seeds);
        Put_Line("Mmoire rinitialise."); 
      else
        -- Le calcul se fait par masque
        for Mask in 0..(Nb_Symbol ** Parameters.Mask_Size)-1 loop
          Put_Line("Traite le masque " & Integer'Image(Mask));
          -- On positionne les bornes pour la premiere sequence
	  Sequences(Zone1.Id).Start := Zone1.Start;
	  Sequences(Zone1.Id).Stop  := Zone1.Stop;
	  Compute_Seeds_By_Mask(Zone1.Id,Zone1.Strand,Mask);
          if Zone2.Id >= Sequences'First and Zone2.Id <= Sequences'Last then
            -- On positionne les bornes pour la seconde squence
	    -- cela se fait ici, car id2 peut etre egal a id1
	    Sequences(Zone2.Id).Start := Zone2.Start;
	    Sequences(Zone2.Id).Stop  := Zone2.Stop;
	    Compute_Seeds_By_Mask(Zone2.Id,Zone2.Strand,Mask);
          end if;
          -- Eval_Seeds n'utilise pas les positions sequences(id).start
	  -- ou stop. Si cela tait modifie, il faut revoir cette partie
	  -- toujours a cause du fait que Id1 peut etre egal a Id2
	  Put_Line("Evaluation des graines pour le masque " & 
                   Integer'Image(Mask));
          Eval_Seeds(Sequences,Seeds,
                     Parameters.Recorded_Length,Parameters.Distance,
                     Parameters.Tolerance,Parameters.Common_Max_Dist,
                     Parameters.Skip,Parameters.Xonly,
                     Zone1,Zone2,Z1_Xonly,Z2_Xonly);
          Put_Line("Graines values pour le masque " & 
                   Integer'Image(Mask));
          Reset_Seeds(Seeds);
          Put_Line("Mmoire rinitialise."); 
        end loop;
      end if;
      -- Transmettre  la base de donnes les nouvelles rgions
      -- et purger la base
      Send_Pairs(Parameters.Final_Length);
      -- Informer le dispatcher que c'est fini
      Dispatcher_Access.Lock;
      Dassirc_Dispatcher.Region_Completed(Client);
      Dispatcher_Access.Unlock;
      Put_Line("Analyse de la rgion termine.");
      Work.Wait(Region_Computing);
    end loop; 
    Cpu_Counters.Remove; 
    Put_Line("Terminaison de la tache: Region_Computing_Task");   
  exception
    when NO_REGION_AVAILABLE =>
      Put_Line("Fin de traitement des rgions");
      Free_Seeds(Seeds);
      Assirc_Seeds.D_Eval_M1.Reset_Pairs_To_Check;
      Work.Terminated(Region_Computing);
      Cpu_Counters.Remove;
      Put_Line("Terminaison de la tache: Region_Computing_Task");   
    when NO_TREATMENT =>
      Put_Line("Dsactivation de la tache de recherche par zone");
      Work.Terminated(Region_Computing);
      Cpu_Counters.Remove; 
      Put_Line("Terminaison de la tache: Region_Computing_Task");   
    when The_Error : others =>
      Put_Line("Exception raised in Region_Computing_Task");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      Work.Terminated(Region_Computing);
      Cpu_Counters.Remove;     
  end Region_Computing_Task;

--------------------------------------------------------------------------
-- La tache de validation de l'alignement                               --
--------------------------------------------------------------------------
  task Alignment_Computing_Task is
    entry Start;
    entry Shutdown;
  end Alignment_Computing_Task;

  task body Alignment_Computing_Task is
    use Assirc_Statistics;
    Max_Putative_Pairs      : constant Integer := 50;
    Max_Alignment_Summaries : constant Integer := 50;
    Putative_Pairs   : Segment_Pairs(1..Max_Putative_Pairs);
    Last    : Integer;
    Summaries : Alignment_Summaries_Array(1..Max_Alignment_Summaries);
    Current : Integer := Summaries'First;
    NO_TREATMENT : exception;
  begin
    Cpu_Counters.Add("Alignment_computing_Task");
    Work.Started(Alignment_Computing);
    Put_Line(" Alignment_Computing_Task : Initialisation de la tache");
    select 
      accept Start;
    or
      accept Shutdown;
      raise NO_TREATMENT;
    or 
      terminate;
    end select;
    Put_Line("Alignment_Computing_Task : Activation de la tache");       
    -- Informer le dispatcher que l'on est pret a aligner
    Dispatcher_Access.Lock;
    Dassirc_Dispatcher.Alignment_Computing_Start(Client);
    Dispatcher_Access.Unlock;
    Put_Line("Alignment_Computing_Task : Declaration au server");       
    Work.Run(Alignment_Computing);
    Put_Line("Alignment_Computing_Task : Running");       
    if Parameters.Alignment then
      loop
	-- Obtenir les paires suivantes
	Database_Access.Lock;
	Dassirc_Database.Next_Pairs(Parameters.Final_Length,Putative_Pairs,Last);
	Database_Access.Unlock;
	exit when Last = 0;
	for I in 1..Last loop
	  Perform_One_Alignment(Sequences,
				Putative_Pairs(I),
				Summaries(Current),
				Parameters.Method,
				Parameters.Max_Length,
				Parameters.Adjust,
				Parameters.Match_Weight,
				Parameters.Mismatch_Weight,
				Parameters.Gap_Weight,
				Parameters.Gap_Length_Weight,
				Parameters.With_End_Weight,
			        Parameters.Similarity_Threshold,
				Parameters.Randomization,
				Parameters.Preserve);
	  if not (Summaries(Current).Start1 = Summaries(Current).Start2 and
		  Summaries(Current).Stop1 = Summaries(Current).Stop2 and
		  Summaries(Current).Name1 = Summaries(Current).Name2)
	    and 
	    (Summaries(Current).Length = 0
	     or 
	     (Summaries(Current).Length >= Parameters.Final_Length and 
	      Summaries(Current).Identity >= Parameters.Minimal_Similarity))
	  then
	    -- On ne considre que le cas o le pourcentage de similarit
	    -- et la longueur aligne dpasse le seuil fix  et le cas
	    -- ou l'alignement n'a pu etre ralis.
	    Statistics.Increment_Validated;
	    if 	Current = Summaries'Last then
	      -- Le buffer est plein, on le poste
	      Dispatcher_Access.Lock;
	      Dassirc_Dispatcher.Append_Summaries(Summaries);
	      Dispatcher_Access.Unlock;
	      Current := Summaries'First;
	    else
	      -- Il reste de la place
	      Current := Current + 1;
	    end if;
	  end if;
	end loop;
      end loop;
      -- Poster le dernier buffer
      if Current > Summaries'First then
	-- On poste le solde
	Dispatcher_Access.Lock;
	Dassirc_Dispatcher.Append_Summaries(Summaries(Summaries'First..Current-1));
	Dispatcher_Access.Unlock;
      end if;
    end if;
    -- Et dire que l'on a fini
    Work.Wait(Alignment_Computing);
    Work.Terminated(Alignment_Computing);
    Dispatcher_Access.Lock;
    Dassirc_Dispatcher.Alignment_Computing_stop(Client);
    Dispatcher_Access.Unlock;
    Cpu_Counters.Remove;
    Put_Line("Terminaison de la tache: Perform_Alignment_Task");    
  exception
    when NO_TREATMENT =>
      Put_Line("Dsactivation de la tache d'alignement");
      Work.Wait(Alignment_Computing);
      Work.Terminated(Alignment_Computing);
      Cpu_Counters.Remove; 
      Put_Line("Terminaison de la tache: Alignment_Computing_Task");   
    when The_Error : others =>
      Put_Line("Exception raised in Alignment_Computing_Task");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      Work.Wait(Alignment_Computing);
      Work.Terminated(Alignment_Computing);
      Cpu_Counters.Remove;
  end Alignment_Computing_Task;

--------------------------------------------------------------------------
-- Une tache informant le dispatcher de l'tat de la base               --
--------------------------------------------------------------------------
  task For_Informing_Dispatcher is
    entry Start;
    entry Complete;
    entry Stop;
  end For_Informing_Dispatcher;

  task body For_Informing_Dispatcher is
    use Ada.Calendar;
    use Assirc_Statistics;
    Period : constant Duration := 2.0; 
    -- Temps entre deux mise a jour du dispatcher 
    Next_Time : Time;
    State     : Partition_State; -- pour connaitre l'etat d'avancement
    Continue : Boolean := True;
  begin
    Cpu_Counters.Add("For_Informing_Dispatcher"); 
    accept Start;
    Next_Time := Clock + Period;
    while Continue loop
      -- regarde si toutes les taches sont finies et si oui, on arrete
      -- sinon on attend une periode. Lors de la rception du signal de
      -- fin, on execute une dernire fois la boucle pour envoyer les 
      -- dernires informations au dispatcher.
      -- Version avec select introduisant un blocage dans la transmission
      -- des infos vers le dispatcher
      -- select
      --   Work.Nothing;
      --   Continue := False;
	-- on quitte la tache si le job est termin
      -- else
	-- on attend la prochaine periode de mise a jour
      --  delay until Next_Time;
      -- end select;
      if Work.Non_Active_Task then
	-- il n'y a plus de tache active, on quitte au prochain
	-- tour seulement pour transmettre une derniere fois les infos
	Continue := False;
      else
	-- On attend jusqu'a la prochjaine echeance
	delay until Next_Time;
      end if;
      Next_Time := Next_Time + Period;
      -- Rcuprer les statistiques
      Statistics.Get_Statistics(State.Constituted_Count,
                                State.Tested_Count,
                                State.Registered_Count,
				State.Validated_Count);
      State.Status := Work.Get_Status;
      --Cpu_Counters.Print_Process_List;
      Cpu_Counters.Elapsed_Times(State.Times);
      Dispatcher_Access.Lock;
      Dassirc_Statistics.Advancement(Client,State);
      Dispatcher_Access.Unlock;
      exit when State.Status = Terminated;
    end loop;
    accept Complete;
    Print_Times;
    accept Stop;
    Cpu_Counters.Remove; 
  exception
    when NOT_CONNECTED =>
      -- Il y a eu deconnexion du client
      Put_Line("Exception NOT_CONNECTED raised in For_Informing_Dispatcher");
      Cpu_Counters.Remove;
      raise Client_Aborted;
    when The_Error : others =>
      Put_Line("Exception raised in For_Informing_Dispatcher");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      Cpu_Counters.Remove;
  end For_Informing_Dispatcher; 
 
--------------------------------------------------------------------------
-- Et l'enchainement de toutes ces taches                               --
--------------------------------------------------------------------------
begin
  -- Enregistrer le process Main
  Cpu_Counters.Add("Main"); 
  -- Et noter dans un fichier le numero du process
  Write_Pid("dassirc_builder");
  -- Lire les options passes au client pour en dduire les actions
  -- acceptable par le client
  Get_Options(Options);
  -- A partir de maintenant, on reserve l'acces au dispatcher
  -- pour cette tache principale qui va rcuperer les donnes et 
  -- les paramtres de traitement.
  Dispatcher_Access.Lock;
  -- dclarer l'existence du client au serveur (ici pour rcuprer
  -- l'ventuelle erreur
  Put_Line("Client registering");
  Dassirc_Dispatcher.Client_Registering(Client,
					Options.Supported_Actions,
					Options.Computing_Method);
  Message(Client,1001);
  -- rcuperer les sequences 
  Sequences := Load_Sequences;
  Message(Client,1003);
  -- rcuprer les paramtres du traitement
  Parameters := Dassirc_Dispatcher.Get_Parameters;
  -- sans oublier le shift que l'on positionne dans la foulee
  Set_Seed_Parameters(Dassirc_Dispatcher.Get_Searching_Shift);
  Message(Client,1004);
  -- On libere l'acces au dispatcher
  Dispatcher_Access.Unlock;
  -- activer les taches en fonction des actions supportes
  if Options.Supported_Actions(Region_Computing) then
    Region_Computing_Task.Start;
  else
    Region_Computing_Task.Shutdown;
  end if;
  if Options.Supported_Actions(Seed_Computing) then
    Put_Line("Activate Seed_Computing_Task");
    Seed_Computing_Task.Start(Options.Computing_Method);
    Put_Line("Activated Seed_Computing_Task");
  else
    -- On n'active pas la tache de recherche de graines
    Seed_Computing_Task.Shutdown;
    -- et donc on ne doit pas gnrer de base de donnes locale en
    -- mthode de calsul simple
    Assirc_Seeds.Database.Close_Local_Database;
  end if;
  if Options.Supported_Actions(Expand_Computing) then
    Put_Line("Avant Expand_Computing_Start for " & Client_Id'Image(Client));
    Dispatcher_Access.Lock;
    Dassirc_Dispatcher.Expand_Computing_Start(Client);
    Dispatcher_Access.Unlock;
    Put_Line("Activate Expand_Computing_Task");
    Run_Of_Expand_Computing := 
      new Expand_Computing_Array(1..Options.Expand_Task);
    Put_Line("Activated Expand_Computing_Task");
    for I in Run_Of_Expand_Computing'Range loop
      Run_Of_Expand_Computing(I).Start;
    end loop;
    Put_Line("Started Expand_Computing_Task");
  end if;
  -- activer l'information du dispatcher
  For_Informing_Dispatcher.Start;
  -- Mettre a jour la liste des processus (rcupration des fils)
  Cpu_Counters.Update_Childs;
  -- attendre que les taches de recherche soient finies 
  Put_Line("Main: attend Work.End_For_Searching");
  Work.End_For_Searching;
  Put_Line("Main: Work.End_Sor_Searching valide");
  -- et si ncessaire informer que la tache d'extension est termine
  if Options.Supported_Actions(Expand_Computing) then
    Dispatcher_Access.Lock;
    Dassirc_Dispatcher.Expand_Computing_Stop(Client);
    Dispatcher_Access.Unlock;
  end if;
  -- Lancer l'alignement
  Alignment_Computing_Task.Start;
  -- Mettre a jour la liste des processus (rcupration des fils)
  Cpu_Counters.Update_Childs;
  -- attendre que les taches soient finies 
  Put_Line("Main: attend Work.Nothing");
  Work.Nothing;
  Put_Line("Main: Work.Nothing valide");
  -- Arret de la tache d'information
  Put_Line("Main: attend For_Informing_Dispatcher.Complete");
  For_Informing_Dispatcher.Complete;
  Put_Line("Main: attend For_Informing_Dispatcher.Stop");
  For_Informing_Dispatcher.Stop;
  -- dtruire la tache For_Informing_Dispatcher
  --abort For_Informing_Dispatcher;
  -- dconnexion du client
  Put_Line("Main: Deconnexion du client");
  Dassirc_Dispatcher.Client_Disconnecting(Client);
  Put_Line("Main: Client deconnecte");
  Message(Client,1002);
  Put_Line("Main: Termin");
  Cpu_Counters.Remove;
  Cpu_Counters.Print_Elapsed_Times;
  -- Fin
  Put_Line("Terminaison du programme principal");        
exception
  when TOO_MANY_CLIENTS =>
    Message(Client,1);
    Cpu_Counters.Remove;
    Print_Times;
    Put_Line("Terminaison du programme (Max client atteint");
  when SEQUENCES_NOT_LOADED =>
    Message(Client,2);
    -- Deconnection du serveur
    Dassirc_Dispatcher.Client_Aborting(Client);
    Dispatcher_Access.Unlock;
    -- Arret des taches locales de traitement
    Seed_Computing_Task.Shutdown;
    Region_Computing_Task.Shutdown;
    Alignment_Computing_Task.Shutdown;
    if Run_Of_Expand_Computing /= null then
      for I in Run_Of_Expand_Computing'Range loop
	abort Run_Of_Expand_Computing(I);
      end loop;
    end if;
    -- Arret de la tache d'information
    Put_Line("Main: attend For_Informing_Dispatcher.Complete");
    abort For_Informing_Dispatcher;
    --For_Informing_Dispatcher.Complete;
    --Put_Line("Main: attend For_Informing_Dispatcher.Stop");
    --For_Informing_Dispatcher.Stop;
    -- Arret du compteur
    Cpu_Counters.Remove;
    Print_Times;
    Put_Line("Terminaison du programme: impossible de charger les sequences");
  when CLIENT_ABORTED =>
    Cpu_Counters.Remove;
    -- Arret des taches locales de traitement
    Seed_Computing_Task.Shutdown;
    Region_Computing_Task.Shutdown;
    Alignment_Computing_Task.Shutdown;
    if Run_Of_Expand_Computing /= null then
      for I in Run_Of_Expand_Computing'Range loop
	abort Run_Of_Expand_Computing(I);
      end loop;
    end if;
    -- Information
    Print_Times;
    Put_Line("Terminaison du programme: deconnexion du client");
  --when CONSTRAINT_ERROR =>
  --  Message(Client,999);
end Dassirc_Builder;
